﻿-- ********************************************************************************
-- Broker Statistics
-- A Data Broker display for combat statistics.
-- By: Shenton
--
-- Core.lua
-- Ace3 libs, core functions, Ace3 init.
-- ********************************************************************************

-- Ace libs (<3)
local A = LibStub("AceAddon-3.0"):NewAddon("BrokerStats", "AceConsole-3.0", "AceTimer-3.0", "AceComm-3.0", "AceEvent-3.0");
A.L = LibStub("AceLocale-3.0"):GetLocale("BrokerStats");

-- LibDataBroker
A.LibDataBroker = LibStub("LibDataBroker-1.1");

-- LibShaedMedia
A.LibSharedMedia = LibStub("LibSharedMedia-3.0");

-- Addon global
_G["DataBrokerStatisticsGlobal"] = A;

-- LUA globals
local mfloor, pairs, ipairs, loadstring = math.floor, pairs, ipairs, loadstring;
local smatch, sgsub, tostring, type = string.match, string.gsub, tostring, type;
local tinsert, tconcat, format, tsort = table.insert, table.concat, string.format, table.sort;
local ssub = string.sub;

-- WoW globals
local strtrim = strtrim;

-- ********************************************************************************
-- Functions
-- ********************************************************************************

--- Send a message to the chat frame with the addon name colored
-- @param text The message to display
-- @param level Verbosity level (0 to 2)
-- @param color Bool, if true will color in red
function A:Message(text, level, color)
    if ( not level ) then level = 1; end

    if ( level <= A.db.profile.verbosity ) then
        if ( color ) then
            color = A.colorRed;
        else
            color = A.colorGreen;
        end

        DEFAULT_CHAT_FRAME:AddMessage(color..A.L["Broker Statistics"]..": "..A.colorReset..text);
    end
end

--- Handle the slash command
function A:SlashCommand(...)
    local arg1 = ...;

    if ( not arg1 or arg1 == "" ) then
        A:OpenConfigPanel();
    elseif ( arg1 and arg1 == "resetprofile" ) then
        StaticPopup_Show("BrokerStatsResetProfile");
    else
        A:Message(A.L["With no argument it will display the configuration panel."], 0);
        A:Message(A.L["   |cFFC79C6Eresetprofile|r - Will reset the current configuration profile."], 0);
    end
end

--- Load config addon and remove config loader from Blizzard options frame
function A:LoadAddonConfig()
    local loaded, reason = LoadAddOn("Broker_StatsConfig");

    if ( loaded ) then
        for k,v in ipairs(INTERFACEOPTIONS_ADDONCATEGORIES) do
            if ( v.name == A.L["Broker Statistics config loader"] ) then
                INTERFACEOPTIONS_ADDONCATEGORIES[k] = nil;
            end
        end
    elseif ( reason ) then
        reason = _G["ADDON_"..reason];
        A:Message(A.L["Failed to load configuration, reason: %s."]:format(reason), 1, 1);
    end

    return loaded;
end

--- Add to blizzard options frame a temporary category
function A:AddToBlizzTemp()
    local f  = CreateFrame("Frame");
    f.name = A.L["Broker Statistics config loader"];

    local b = CreateFrame("Button", nil, f, "UIPanelButtonTemplate");
    b:SetSize(140, 22);
    b:SetPoint("TOPLEFT", f, "TOPLEFT", 20, -20);
    b:SetText(A.L["Load configuration"]);
    b:SetScript("OnClick", function(self)
        local loaded = A:LoadAddonConfig();

        if ( loaded ) then
            InterfaceAddOnsList_Update();
            InterfaceOptionsFrame_OpenToCategory(A.L["Broker Statistics"]);
        end
    end);

    InterfaceOptions_AddCategory(f);
end

--- Display configuration panel
-- Load it if needed
function A:OpenConfigPanel()
    if ( A.AceConfigDialog ) then
        A.AceConfigDialog:Open("BrokerStatsConfig");
    else
        local loaded = A:LoadAddonConfig();

        if ( loaded ) then
            A.AceConfigDialog:Open("BrokerStatsConfig");
        end
    end
end

--- Round function (http://lua-users.org/wiki/SimpleRound)
-- @param num The number to round
-- @param idp Decimal places
function A:Round(num, idp)
    if ( type(num) == "number" ) then
        local mult = 10^(idp or 0);
        return mfloor(num * mult + 0.5) / mult;
    else
        A:Message(A.L["Attempt to perform arithmetic on a non number value, you are probably trying to round a string."], 1, 1);
        return num;
    end
end

--- Color percent to hex
local b16 =
{
    [0] = "0",
    [1] = "1",
    [2] = "2",
    [3] = "3",
    [4] = "4",
    [5] = "5",
    [6] = "6",
    [7] = "7",
    [8] = "8",
    [9] = "9",
    [10] = "a",
    [11] = "b",
    [12] = "c",
    [13] = "d",
    [14] = "e",
    [15] = "f"
};
function A:PercentToHex(r, g, b, a)
    r = A:Round(r * 255);
    g = A:Round(g * 255);
    b = A:Round(b * 255);
    a = A:Round(a * 255);

    r = b16[mfloor(r / 16)]..b16[r % 16];
    g = b16[mfloor(g / 16)]..b16[g % 16];
    b = b16[mfloor(b / 16)]..b16[b % 16];
    a = b16[mfloor(a / 16)]..b16[a % 16];

    return a..r..g..b;
end

--- Convert table to string (http://lua-users.org/wiki/TableUtils)
function A:TableValToStr(v)
    if ( "string" == type(v) ) then
        v = sgsub(v, "\n", "\\n");
        if ( smatch(sgsub(v, "[^'\"]", ""), '^"+$') ) then return "'"..v.."'"; end
        return '"'..sgsub(v, '"', '\\"')..'"';
    else
        return "table" == type(v) and A:TableTostring(v) or tostring(v);
    end
end
function A:TableKeyToStr(k)
    if ( "string" == type(k) and smatch(k, "^[_%a][_%a%d]*$") ) then return k; end

    return "["..A:TableValToStr( k ) .. "]";
end
function A:TableTostring(tbl)
    local result, done = {}, {};

    for k,v in ipairs(tbl) do
        tinsert(result, A:TableValToStr(v));
        done[k] = true;
    end

    for k,v in pairs(tbl) do
        if ( not done[k] ) then tinsert(result, A:TableKeyToStr(k).."="..A:TableValToStr(v)); end
    end

    return "{"..tconcat(result, ",").."}";
end

--- Pairs function with alphabetic sort
function A:PairsByKeys(t, f)
    local a, i = {}, 0;

    for n in pairs(t) do a[#a+1] = n; end
    tsort(a, f);

    local iter = function()
        i = i + 1;
        if ( not a[i] ) then
            return nil;
        else
            return a[i], t[a[i]];
        end
    end

    return iter;
end

--- Simple shallow copy for copying specialization profiles
-- Shamelessly ripped off from Ace3 AceDB-3.0
-- Did I say I love you guys? :p
function A:CopyTable(src, dest)
	if ( type(dest) ~= "table" ) then dest = {}; end

	if ( type(src) == "table" ) then
		for k,v in pairs(src) do
			if ( type(v) == "table" ) then
				v = A:CopyTable(v, dest[k]);
			end

			dest[k] = v;
		end
	end

	return dest;
end

--- Create the needed (empty) tables within the config db if needed
function A:CreateNeededConfigTables()
    if ( not A.db.profile.defaultDisplay[A.playerSpec] ) then
        A.db.profile.defaultDisplay[A.playerSpec] =
        {
            displayPre = 1,
            displayPost = 1,
            statsSeparator = "||",
            icon = A.defaultLdbIcon,
            statIconSize = 16,
            preColor =
            {
                r = 0,
                g = 0.5,
                b = 0.5,
                a = 1
            },
            preColorHex = "|cff008080",
            postColor =
            {
                r = 1,
                g = 1,
                b = 1,
                a = 1
            },
            postColorHex = "|cffffffff",
            separatorColor =
            {
                r = 1,
                g = 1,
                b = 1,
                a = 1
            },
            separatorColorHex = "|cffffffff"
        };
    end
    if ( not A.db.profile.dataBroker.customDisplays[A.playerSpec] ) then A.db.profile.dataBroker.customDisplays[A.playerSpec] = {}; end
    if ( not A.db.profile.builtInStats[A.playerSpec] ) then A.db.profile.builtInStats[A.playerSpec] = {}; end
    if ( not A.db.profile.custom[A.playerSpec] ) then A.db.profile.custom[A.playerSpec] = {}; end
    if ( not A.db.profile.specProfiles[A.playerSpec] ) then A.db.profile.specProfiles[A.playerSpec] = {}; end
end

--- Remove unused entries from the database
function A:DelUnusedEntry()
    for k,v in pairs(A.db.profile.dataBroker.customDisplays) do
        for kk,vv in pairs(v) do
            if ( not A.db.global.customDisplays[kk] ) then
                A.db.profile.dataBroker.customDisplays[k][kk] = nil;
            end
        end
    end

    for k in pairs(A.db.profile.builtInStats) do
        for kk,vv in pairs(A.db.profile.builtInStats[k]) do
            if ( vv.dbObject ) then
                local ok;

                for kkk in pairs(A.db.global.customDisplays) do
                    if ( kkk == vv.dbObject ) then ok = 1; end
                end

                if ( not ok ) then
                    A.db.profile.builtInStats[k][kk].dbObject = "Default";
                end
            end

            if ( vv.alert ) then
                local ok;

                for kkk in pairs(A.db.global.alerts) do
                    if ( kkk == vv.alert ) then ok = 1; end
                end

                if ( not ok ) then
                    A.db.profile.builtInStats[k][kk].alert = "none";
                end
            end
        end
    end

    for k in pairs(A.db.profile.custom) do
        for kk,vv in pairs(A.db.profile.custom[k]) do
            if ( not A.db.global.custom[kk] ) then
                A.db.profile.custom[k][kk] = nil;
            else
                if ( vv.dbObject ) then
                    local ok;

                    for kkk in pairs(A.db.global.customDisplays) do
                        if ( kkk == vv.dbObject ) then ok = 1; end
                    end

                    if ( not ok ) then
                        A.db.profile.custom[k][kk].dbObject = "Default";
                    end
                end

                if ( vv.alert ) then
                    local ok;

                    for kkk in pairs(A.db.global.alerts) do
                        if ( kkk == vv.alert ) then ok = 1; end
                    end

                    if ( not ok ) then
                        A.db.profile.custom[k][kk].alert = "none";
                    end
                end
            end
        end
    end
end

-- Return if there is a least one custom stat
function A:GotCustom()
    for _,v in pairs(A.db.global.custom) do
        if ( v.calc and v.calc ~= "" ) then return 1; end
    end

    return nil;
end

-- Return if there is a least one custom ldb object
function A:GotCustomLdb()
    for k in pairs(A.db.global.customDisplays) do
        if ( k ) then return 1; end
    end

    return nil;
end

-- Return if there is a least one alert
-- function A:GotAlert()
    -- for k in pairs(A.db.global.alerts) do
        -- if ( k ) then return 1; end
    -- end

    -- return nil;
-- end

--- Create the select table for ldb objects
local out;
function A:CreateDbObjectsSelect()
    out =
    {
        [A.L["Default"]] = A.L["Default"];
    };

    for k in A:PairsByKeys(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        out[k] = k;
    end

    return out;
end

--- Create the select table for specialization profiles
function A:CreateSpecProfilesSelect(noDefault)
    out = {};

    for k in A:PairsByKeys(A.db.profile.builtInStats) do
        if ( noDefault and k == A.L["Default"] ) then
        else
            out[k] = k;
        end
    end

    return out;
end

--- Check if the alert is used
function A:IsAlertInUse(name)
    for k,v in pairs(A.db.profile.builtInStats[A.playerSpec]) do
        if ( v.alert == name ) then return k; end
    end

    for k,v in pairs(A.db.profile.custom[A.playerSpec]) do
        if ( v.alert == name ) then return k; end
    end

    return nil;
end

--- Create the select table for alerts
function A:CreateAlertsSelect(current)
    out =
    {
        none = A.L["None"];
    };

    for k in A:PairsByKeys(A.db.global.alerts) do
        if ( not A:IsAlertInUse(k) or current == k ) then
            out[k] = k;
        end
    end

    return out;
end

--- Create the select table for order
function A:CreateOrderSelect()
    local index = 0;
    out = {};

    for _ in pairs(A.builtInStats) do
        index = index + 1;
        out[index] = index;
    end

    for _ in pairs(A.db.global.custom) do
        index = index + 1;
        out[index] = index;
    end

    return out;
end

--- Create order description table
function A:CreateOrderDescriptions()
    out = {};

    for k,v in A:PairsByKeys(A.db.profile.builtInStats[A.playerSpec]) do
        if ( v.enabled and v.order ) then
            local dbObject;

            if ( not v.dbObject ) then
                dbObject = A.L["Default"];
            else
                dbObject = v.dbObject
            end

            if ( not out[dbObject] ) then out[dbObject] = {}; end

            out[dbObject][#out[dbObject]+1] =
            {
                name = k,
                order = v.order,
            };
        end
    end

    for k,v in A:PairsByKeys(A.db.profile.custom[A.playerSpec]) do
        if ( v.enabled and v.order ) then
            local dbObject;

            if ( not v.dbObject ) then
                dbObject = A.L["Default"];
            else
                dbObject = v.dbObject
            end

            if ( not out[dbObject] ) then out[dbObject] = {}; end

            out[dbObject][#out[dbObject]+1] =
            {
                name = k,
                order = v.order,
            };
        end
    end

    for k,v in pairs(out) do
        local str = "";

        tsort(v, function(a, b) return a.order < b.order; end);

        for _,vv in ipairs(v) do
            str = str..vv.name..": |cffffd100"..vv.order.."|r, ";
        end

        out[k] = strtrim(str, ", ");
    end

    return out;
end

--- Return informations about the current specialization
function A:GetSpecInfos()
    local specGroup = GetActiveSpecGroup(false);
    local maxNumTalents, column, selected, numOut = 18, nil, nil, "";
    local _, nameOut = GetSpecializationInfo(specGroup, false);

    for i=1,maxNumTalents do
        _, _, _, column, selected = GetTalentInfo(i);

        if ( selected ) then
            numOut = numOut.."/"..column;
        end
    end

    if ( numOut == "") then
        numOut = "("..A.L["No Talents"]..")";
    else
        numOut = "("..strtrim(numOut, "/")..")";
    end

    return nameOut, numOut;
end

--- Get the current default spec var
function A:GetSpecVar()
    local name, num = A:GetSpecInfos();

    if ( name and A.db.profile.configAccordingSpec ) then
        if ( A.db.profile.configAccordingSpecBehavior == 2 ) then
            return name.." "..num;
        else
            return name;
        end
    else
        return A.L["Default"];
    end
end

--- Set the current spec var
function A:SetSpecVar()
    local spec = A:GetSpecVar();

    if ( A.db.profile.specProfiles[spec] and A.db.profile.specProfiles[spec].useSpec ) then
        A.playerSpec = A.db.profile.specProfiles[spec].useSpec;
    else
        A.playerSpec = spec;
    end

    A:CreateNeededConfigTables();
end

--- Pause updater
function A:Pause()
    A:CancelTimer(A.mainTimer, 1);
    A:SetTextOnLdbObjects(A.colorRed..A.L["Paused"]);
    A.paused = 1;
end

--- Unpause updater
function A:Unpause()
    A.mainTimer = A:ScheduleRepeatingTimer("UpdateText", A.db.profile.time);
    A.paused = nil;
end

--- Pause toggle
function A:PauseToggle()
    if ( A.paused ) then
        A:Unpause();
    else
        A:Pause();
    end
end

--- Refresh the main timer to the current delay
function A:RefreshTimer()
    A.db.profile.time = A:Round(A.db.profile.time, 1);
    A:CancelTimer(A.mainTimer, 1);
    A.mainTimer = A:ScheduleRepeatingTimer("UpdateText", A.db.profile.time);
end

--- Count the number of activated statistics and ldb objects
-- @return the number of stats and objects activated and their names
function A:CountActivated()
    local stats, countStats, objects, countObjects = {}, 0, {}, 0;

    for k,v in A:PairsByKeys(A.db.profile.builtInStats[A.playerSpec]) do
        if ( v.enabled ) then
            stats[#stats+1] = {
                name = k,
                icon = v.icon,
                color = v.colorHex,
            };
            countStats = countStats + 1;
        end
    end

    for k,v in A:PairsByKeys(A.db.profile.custom[A.playerSpec]) do
        if ( v.enabled ) then
            stats[#stats+1] = {
                name = k,
                icon = v.icon,
                color = v.colorHex,
            };
            countStats = countStats + 1;
        end
    end

    for k,v in A:PairsByKeys(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        if ( v.enabled ) then
            objects[#objects+1] = k;
            countObjects = countObjects + 1;
        end
    end

    objects[#objects+1] = A.L["Default"];

    return stats, countStats, objects, countObjects;
end

--- Display a string on all enabled ldb objects
function A:SetTextOnLdbObjects(text)
    A.ldbObjectDefault.text = text;

    for k,v in pairs(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        if ( A["ldbObject"..k] and v.enabled ) then A["ldbObject"..k].text = text; end
    end
end

--- Create the others Data Broker objects
function A:CreateDataBrokerOjbects()
    for k,v in pairs(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        if ( v.enabled ) then
            local icon;

            if ( v.icon ) then
                icon = v.icon;
            else
                icon = A.defaultLdbIcon;
            end

            A:CreateDataBrokerOjbect(k, icon);
        end
    end
end

--- Create a Data Broker objects if needed
function A:CreateDataBrokerOjbect(name, icon)
    if ( not name ) then name = "Default"; end

    if ( A.LibDataBroker:GetDataObjectByName("Broker Stats - "..name) ) then
        A["ldbObject"..name] = A.LibDataBroker:GetDataObjectByName("Broker Stats - "..name);
        A["ldbObject"..name].text = A.L["Nothing to display"];
        A["ldbObject"..name].label = A.L["Broker Statistics"];
        A["ldbObject"..name].icon = icon;
        A["ldbObject"..name].OnClick = function(self, button)
            if (button == "LeftButton") then
                A:PauseToggle();
            elseif ( button == "RightButton" ) then
                A:DropDownMenu(self, name);
            elseif ( button == "MiddleButton" ) then
                A:OpenConfigPanel();
            end
        end;
        A["ldbObject"..name].OnTooltipShow = function(tooltip)
            local statsNames, countStats, objectsName, countObjects = A:CountActivated();
            local stats, objects, count, color, icon = "", "", 0, "", "";

            tooltip:AddDoubleLine(A.colorWhite..A.L["Broker Statistics"], A.colorGreen.."v"..A.version);
            tooltip:AddLine(" ");
            tooltip:AddLine(A.L["Statistics displayed: %d"]:format(countStats));

            for _,v in ipairs(statsNames) do
                if ( v.icon ) then
                    icon = "|T"..v.icon..":16|t ";
                else
                    icon = "|T"..A.noIconIcon..":16|t ";
                end

                if ( v.color ) then
                    color = v.color;
                else
                    color = "";
                end

                tooltip:AddLine(icon..color..v.name);
            end

            tooltip:AddLine(" ");
            tooltip:AddLine(A.L["Data broker objects used: %d"]:format(countObjects + 1));

            for _,v in ipairs(objectsName) do
                tooltip:AddLine(v);
            end

            tooltip:AddLine(" ");
            tooltip:AddLine(A.L["|cFFC79C6ELeft-Click: |cFF33FF99Pause the addon update timer\n|cFFC79C6ERight-Click: |cFF33FF99Open the quick access menu\n|cFFC79C6EMiddle-Click: |cFF33FF99Open the configuration panel"]);
        end;
    else
        A["ldbObject"..name] = A.LibDataBroker:NewDataObject("Broker Stats - "..name,
        {
            type = "data source",
            text = A.L["Nothing to display"],
            label = A.L["Broker Statistics"],
            icon = icon,
            tocname = "Broker_Stats",
            OnClick = function(self, button)
                if (button == "LeftButton") then
                    A:PauseToggle();
                elseif ( button == "RightButton" ) then
                    A:DropDownMenu(self, name);
                elseif ( button == "MiddleButton" ) then
                    A:OpenConfigPanel();
                end
            end,
            OnTooltipShow = function(tooltip)
                local statsNames, countStats, objectsName, countObjects = A:CountActivated();
                local stats, objects, count, color, icon = "", "", 0, "", "";

                tooltip:AddDoubleLine(A.colorWhite..A.L["Broker Statistics"], A.colorGreen.."v"..A.version);
                tooltip:AddLine(" ");
                tooltip:AddLine(A.L["Statistics displayed: %d"]:format(countStats));

                for _,v in ipairs(statsNames) do
                    if ( v.icon ) then
                        icon = "|T"..v.icon..":16|t ";
                    else
                        icon = "|T"..A.noIconIcon..":16|t ";
                    end

                    if ( v.color ) then
                        color = v.color;
                    else
                        color = "";
                    end

                    tooltip:AddLine(icon..color..v.name);
                end

                tooltip:AddLine(" ");
                tooltip:AddLine(A.L["Data broker objects used: %d"]:format(countObjects + 1));

                for _,v in ipairs(objectsName) do
                    tooltip:AddLine(v);
                end

                tooltip:AddLine(" ");
                tooltip:AddLine(A.L["|cFFC79C6ELeft-Click: |cFF33FF99Pause the addon update timer\n|cFFC79C6ERight-Click: |cFF33FF99Open the quick access menu\n|cFFC79C6EMiddle-Click: |cFF33FF99Open the configuration panel"]);
            end,
        });
    end
end

--- Hide disabled ldb objects
function A:HideDisabledLdb()
    for k,v in pairs(A.db.global.customDisplays) do
        if ( A["ldbObject"..k] ) then
            if ( not A.db.profile.dataBroker.customDisplays[A.playerSpec][k] ) then
                A["ldbObject"..k].text = "";
                A["ldbObject"..k].label = "";
                A["ldbObject"..k].icon = "";
                A["ldbObject"..k].OnTooltipShow = nil;
                A["ldbObject"..k].Onclick = nil;
            elseif ( not A.db.profile.dataBroker.customDisplays[A.playerSpec][k].enabled ) then
                A["ldbObject"..k].text = "";
                A["ldbObject"..k].label = "";
                A["ldbObject"..k].icon = "";
                A["ldbObject"..k].OnTooltipShow = nil;
                A["ldbObject"..k].Onclick = nil;
            end
        end
    end
end

--- Order function for sorting update table
local function compare(a, b)
    if ( not a ) then return true; end
    if ( not b ) then return false; end

    local aOrder = a.order;
    local bOrder = b.order;

    if ( aOrder == bOrder ) then
        return a.name < b.name;
    end

    return aOrder < bOrder;
end

--- Create statistics table
local customDisplays;
function A:CreateStatisticsTable()
    customDisplays =
    {
        Default =
        {
            displayPre = A.db.profile.defaultDisplay[A.playerSpec].displayPre,
            displayPost = A.db.profile.defaultDisplay[A.playerSpec].displayPost,
            preColorHex = A.db.profile.defaultDisplay[A.playerSpec].preColorHex,
            postColorHex = A.db.profile.defaultDisplay[A.playerSpec].postColorHex,
            statsSeparator = " "..A.db.profile.defaultDisplay[A.playerSpec].separatorColorHex..A.db.profile.defaultDisplay[A.playerSpec].statsSeparator..A.colorReset.." ",
            statIconSize = A.db.profile.defaultDisplay[A.playerSpec].statIconSize
        }
    };

    for k,v in pairs(A.db.profile.dataBroker.customDisplays[A.playerSpec]) do
        if ( v.enabled ) then
            customDisplays[k] =
            {
                displayPre = v.displayPre,
                displayPost = v.displayPost,
                preColorHex = v.preColorHex,
                postColorHex = v.postColorHex,
                statsSeparator = " "..v.separatorColorHex..v.statsSeparator..A.colorReset.." ",
                statIconSize = v.statIconSize
            };
        end
    end

    A.statsTable = {};

    for k,v in A:PairsByKeys(A.db.profile.builtInStats[A.playerSpec]) do
        if ( not A.builtInStats[k] ) then
            A.db.profile.builtInStats[A.playerSpec][k] = nil;
            A:Message(A.L["Stats %s is no longer available, removing it from your %s profile."]:format(k, A.playerSpec), 0, 1);
        elseif ( v.enabled ) then
            local dbObject = v.dbObject;

            if ( not dbObject ) then dbObject = "Default"; end

            if ( customDisplays[dbObject] ) then
                local index = #A.statsTable + 1;
                local opt = A.builtInStats[k];
                local calc = loadstring(opt.calc);

                A.statsTable[index] =
                {
                    name = k,
                    dbObject = dbObject,
                    multiple = opt.multiple,
                    round = opt.round,
                    separator = opt.separator,
                    calc = calc,
                    statsSeparator = customDisplays[dbObject].statsSeparator
                };

                -- Pre
                if ( v.pre and v.pre ~= "" and customDisplays[dbObject].displayPre ) then
                    A.statsTable[index].pre = customDisplays[dbObject].preColorHex..v.pre..A.colorReset;
                elseif ( opt.pre and customDisplays[dbObject].displayPre ) then
                    A.statsTable[index].pre = customDisplays[dbObject].preColorHex..opt.pre..A.colorReset;
                else
                    A.statsTable[index].pre = "";
                end

                -- Post
                if ( v.post and v.post ~= "" and customDisplays[dbObject].displayPost ) then
                    A.statsTable[index].post = customDisplays[dbObject].postColorHex..v.post..A.colorReset;
                elseif ( opt.post and customDisplays[dbObject].displayPost ) then
                    A.statsTable[index].post = customDisplays[dbObject].postColorHex..opt.post..A.colorReset;
                else
                    A.statsTable[index].post = "";
                end

                -- Icon
                if ( v.iconEnabled ) then
                    if ( v.icon ) then
                        A.statsTable[index].icon = "|T"..v.icon..":"..customDisplays[dbObject].statIconSize.."|t";
                    else
                        A.statsTable[index].icon = "|T"..A.defaultStatIcon..":"..customDisplays[dbObject].statIconSize.."|t";
                    end
                else
                    A.statsTable[index].icon = "";
                end

                -- Stat color
                if ( v.colorHex ) then
                    A.statsTable[index].color = v.colorHex;
                else
                    A.statsTable[index].color = "";
                end

                -- Alert
                if ( v.alert and v.alert ~= "none" and v.alertValue ) then
                    local check = calc();
                    check = type(check);

                    if ( check ~= "number" ) then
                        A:Message(A.L["The alert %s of statistic %s is trying to compare a number with %s."]:format(v.alert, k, check) ,1, 1);
                    else
                        A:SetAlert(v.alert);
                        A.statsTable[index].alert = v.alert;
                        A.statsTable[index].alertValue = v.alertValue;
                        if ( v.alertMode == "proc" ) then
                            A.statsTable[index].alertTimer = v.alertTimer;
                        end
                    end
                end

                -- Order
                if ( v.order ) then
                    A.statsTable[index].order = tonumber(v.order);
                else
                    A.statsTable[index].order = 1000;
                end
            end
        end
    end

    for k,v in A:PairsByKeys(A.db.profile.custom[A.playerSpec]) do
        if ( v.enabled ) then
            local dbObject = v.dbObject;

            if ( not dbObject ) then dbObject = "Default"; end

            if ( customDisplays[dbObject] ) then
                local opt, calc = A.db.global.custom[k], nil;

                if ( opt.calc ) then calc = loadstring(opt.calc); end

                local ok, errorString = pcall(calc);

                if ( ok ) then
                    if ( calc ) then
                        local index = #A.statsTable + 1;

                        A.statsTable[index] =
                        {
                            name = k,
                            dbObject = dbObject,
                            multiple = opt.multiple,
                            round = opt.round,
                            separator = opt.separator,
                            order = v.order,
                            calc = calc,
                            statsSeparator = customDisplays[dbObject].statsSeparator
                        };

                        -- Pre
                        if ( opt.pre and customDisplays[dbObject].displayPre ) then
                            A.statsTable[index].pre = customDisplays[dbObject].preColorHex..opt.pre..A.colorReset;
                        else
                            A.statsTable[index].pre = "";
                        end

                        -- Post
                        if ( opt.post and customDisplays[dbObject].displayPost ) then
                            A.statsTable[index].post = customDisplays[dbObject].postColorHex..opt.post..A.colorReset;
                        else
                            A.statsTable[index].post = "";
                        end

                        -- Icon
                        if ( v.iconEnabled ) then
                            if ( v.icon ) then
                                A.statsTable[index].icon = "|T"..v.icon..":"..customDisplays[dbObject].statIconSize.."|t";
                            else
                                A.statsTable[index].icon = "|T"..A.defaultStatIcon..":"..customDisplays[dbObject].statIconSize.."|t";
                            end
                        else
                            A.statsTable[index].icon = "";
                        end

                        -- Stat color
                        if ( v.colorHex ) then
                            A.statsTable[index].color = v.colorHex;
                        else
                            A.statsTable[index].color = "";
                        end

                        -- Alert
                        if ( v.alert and v.alert ~= "none" and v.alertValue ) then
                            local check = calc();
                            check = type(check);

                            if ( check ~= "number" ) then
                                A:Message(A.L["The alert %s of statistic %s is trying to compare a number with %s."]:format(v.alert, k, check) ,1, 1);
                            else
                                A:SetAlert(v.alert);
                                A.statsTable[index].alert = v.alert;
                                A.statsTable[index].alertValue = v.alertValue;
                                if ( v.alertMode == "proc" ) then
                                    A.statsTable[index].alertTimer = v.alertTimer;
                                end
                            end
                        end

                        -- Order
                        if ( v.order ) then
                            A.statsTable[index].order = tonumber(v.order);
                        else
                            A.statsTable[index].order = 1000;
                        end
                    else
                        A:Message(A.L["There is an error in %s code. Error: %s"]:format(k, A.L["Returned value is nil."]), 1, 1);
                    end
                else
                    A:Message(A.L["There is an error in %s code. Error: %s"]:format(k, errorString), 1, 1);
                end
            end
        end
    end

    tsort(A.statsTable, compare);

    if ( A.paused ) then
        A:SetTextOnLdbObjects(A.colorRed..A.L["Paused"]);
    else
        A:SetTextOnLdbObjects(A.L["Nothing to display"]);
    end
end

--- Update text on the data broker display
local updateText;
function A:UpdateText()
    updateText = {};

    for _,v in ipairs(A.statsTable) do
        local stat = "";

        if ( v.multiple ) then
            local values = { v.calc() };

            for _,vv in pairs(values) do
                if ( v.round ) then
                    stat = stat..A:Round(vv, v.round)..v.post..v.separator;
                else
                    stat = stat..vv..v.post..v.separator;
                end
            end

            stat = strtrim(stat, v.separator);

            if ( not updateText[v.dbObject] ) then
                updateText[v.dbObject] = v.pre..v.icon..v.color..stat..A.colorReset;
            else
                updateText[v.dbObject] = updateText[v.dbObject]..v.statsSeparator.." "..v.pre..v.icon..v.color..stat..A.colorReset;
            end
        else
            if ( v.round ) then
                stat = A:Round(v.calc(), v.round);
            else
                stat = v.calc();
            end

            if ( not updateText[v.dbObject] ) then
                updateText[v.dbObject] = v.pre..v.icon..v.color..stat..A.colorReset..v.post;
            else
                updateText[v.dbObject] = updateText[v.dbObject]..v.statsSeparator..v.pre..v.icon..v.color..stat..A.colorReset..v.post;
            end

            if ( v.alert ) then
                if ( v.alertTimer ) then
                    if ( stat >= v.alertValue ) then
                        A:Alert(v.alert, nil, v.alertTimer);
                    else
                        A:Alert(v.alert, 1, 1);
                    end
                else
                    if ( stat >= v.alertValue ) then
                        A:Alert(v.alert);
                    else
                        A:Alert(v.alert, 1);
                    end
                end
            end
        end
    end

    for k,v in pairs(updateText) do
        if ( A.paused ) then
            v = A:SetTextOnLdbObjects(A.colorRed..A.L["Paused"]);
        elseif ( v == "" ) then v = A.L["Nothing to display"]; end

        A["ldbObject"..k].text = v;
    end
end

-- ********************************************************************************
-- Addon Communication
-- ********************************************************************************

--- Called when receiving an addon communication message
function A:AddonComm(...)
    local prefix, customText, method, who, customName;

    prefix, customText, method, who = ...;

    if ( prefix ~= "BrokerStatsComm" ) then return; end
    if ( who == UnitName("player") ) then return; end

    if ( (method == "WHISPER" and A.db.profile.shareWhisper) or
    (method == "RAID" and A.db.profile.shareRaid) or
    (method == "GUILD" and A.db.profile.shareGuild) or
    (method == "BATTLEGROUND" and A.db.profile.shareBg) ) then
        A:Message(A.L["%s tried to send you a custom stat (%s) with method %s."]:format(who, customName, method), 2, 1);
        return;
    end

    customName = smatch(customText, "^(.+){");
    customText = smatch(customText, "{(.+)}");
    customText = "{"..customText.."}";

    if ( A.db.global.custom[customName] ) then
        local pop = StaticPopup_Show("BrokerStatsReceiveCustomNameExists", customName, who);
        if (pop) then pop.customText = customText; end
    elseif ( A.builtInStats[customName] ) then
        local pop = StaticPopup_Show("BrokerStatsReceiveCustomNameExists", customName, who);
        if (pop) then pop.customText = customText; end
    else
        local pop = StaticPopup_Show("BrokerStatsReceiveCustom", customName, who);
        if (pop) then
            pop.customText = customText;
            pop.customName = customName;
        end
    end
end

--- Send an addon communication message with a custom stat
-- @param target The message target
-- @param customStat The custom stat name
function A:SendStat(target, customStat)
    if ( not target or not customStat ) then return; end

    target = strtrim(target);
    if ( target == "" ) then return; end

    local textMessage = A:TableTostring(A.db.global.custom[customStat]);

    if ( target == "RAID" or target == "GUILD" or target == "BATTLEGROUND" ) then
        A:SendCommMessage("BrokerStatsComm", customStat..textMessage, target);
    else
        A:SendCommMessage("BrokerStatsComm", customStat..textMessage, "WHISPER", target);
    end
end

-- ********************************************************************************
-- Callbacks
-- ********************************************************************************

--- Called when an action is made on a configuration profile
function A:RefreshConfig()
    A:SetSpecVar();
    if ( A.AceConfigRegistry ) then A.AceConfigRegistry:NotifyChange("BrokerStatsConfig"); end
    A:RefreshTimer();
    A:CreateDataBrokerOjbects();
    A:HideDisabledLdb();
    A:CreateStatisticsTable();
end

--- Callback fo event PLAYER_TALENT_UPDATE
function A:PLAYER_TALENT_UPDATE()
    A:RefreshConfig();
end

-- ********************************************************************************
-- Shared Media
-- ********************************************************************************

--- Add sounds to shared media
function A:AddSoundsToSharedMedia()
    for _,v in pairs(A.alertSounds) do
        local name = smatch(v, "\\([%w_%-]+).ogg$");

        A.LibSharedMedia:Register("sound", name, v);
    end
end

-- ********************************************************************************
-- Ace default database
-- ********************************************************************************

--- Default database table
function A:AceDefaultDB()
    local db =
    {
        profile =
        {
            verbosity = 1,
            time = 1,
            shareWhisper = 1,
            shareGuild = 1,
            shareRaid = 1,
            shareBg = 1,
            configAccordingSpecBehavior = 1,
            defaultDisplay = {},
            builtInStats = {},
            custom = {},
            specProfiles = {},
            dataBroker =
            {
                defaultObject = {},
                customDisplays = {},
            },
        },
        global =
        {
            custom = {},
            customDisplays = {},
            alerts = {},
        },
    };

    return db;
end

-- ********************************************************************************
-- Main
-- ********************************************************************************

--- AceAddon callback
-- Called after the addon is fully loaded
function A:OnInitialize()
    -- Database
    A.db = LibStub("AceDB-3.0"):New("BrokerStatsDB", A:AceDefaultDB());

    -- Database revision check
    if ( not A.db.global.databaseRevision ) then
        A.db.global.databaseRevision = A.databaseRevision;
    elseif ( A.db.global.databaseRevision < A.databaseRevision ) then
        A:SetEnabledState(false);
        -- do some stuff on the database
    end

    if A:IsEnabled() then
        -- Config settings refresh
        A.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig");
        A.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig");
        A.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig");

        -- Slash command
        A:RegisterChatCommand("brokerstats", "SlashCommand");
        A:RegisterChatCommand("stats", "SlashCommand");
        A:RegisterChatCommand("bs", "SlashCommand");
    end
end

--- AceAddon callback
-- Called during the PLAYER_LOGIN event
function A:OnEnable()
    -- Default LDB object
    A:CreateDataBrokerOjbect(nil, A.db.profile.dataBroker.defaultObject.icon or A.defaultLdbIcon);

    -- DropDownMenu frame
    A.dropDownMenuFrame = CreateFrame("Frame", "BrokerStatisticsMenuFrame");
    A.dropDownMenuFrame.displayMode = "MENU";

    -- Create "Default" config tables on the fisrt load
    A:CreateNeededConfigTables();

    -- Add sounds to shared media
    A:AddSoundsToSharedMedia();

    -- Delete unused database entries
    A:DelUnusedEntry();

    -- Events
    A:RegisterEvent("PLAYER_TALENT_UPDATE");

    -- Register for addon communication
    A:RegisterComm("BrokerStatsComm", "AddonComm");

    -- Add to blizzard options frame a temporary category
    A:AddToBlizzTemp();

    -- Init
    A:RefreshConfig();
end

--- AceAddon callback
-- Called when the addon is manually disabled
function A:OnDisable()
    A:CancelTimer(A.mainTimer, 1);
    A:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
end
